#include <tests/llvm.h>
const char test_llvm__bpf_test_kbuild_prog[] =
"// SPDX-License-Identifier: GPL-2.0\n"
"/*\n"
" * bpf-script-test-kbuild.c\n"
" * Test include from kernel header\n"
" */\n"
"#ifndef LINUX_VERSION_CODE\n"
"# error Need LINUX_VERSION_CODE\n"
"# error Example: for 4.2 kernel, put 'clang-opt=\"-DLINUX_VERSION_CODE=0x40200\" into llvm section of ~/.perfconfig'\n"
"#endif\n"
"#define SEC(NAME) __attribute__((section(NAME), used))\n"
"\n"
"#include <uapi/linux/fs.h>\n"
"\n"
"SEC(\"func=vfs_llseek\")\n"
"int bpf_func__vfs_llseek(void *ctx)\n"
"{\n"
"	return 0;\n"
"}\n"
"\n"
"char _license[] SEC(\"license\") = \"GPL\";\n"
"int _version SEC(\"version\") = LINUX_VERSION_CODE;\n"
;
